/**
 * @file main.c
 * @author your name (you@domain.com)
 * @brief 
 * @version 0.1
 * @date 2019-10-15
 * 
 * @copyright Copyright (c) 2019 
 * 
 */
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <mpi.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>

typedef struct timeval timeval_t;
typedef struct timespec timespec_t;

int rank, size;

timespec_t start, end;

int msg_size_start, msg_size_end;
#define MSG_SIZE_START (msg_size_start)
#define MSG_SIZE_END (msg_size_end) 

int msg_count_start, msg_count_end;
#define MSG_COUNT_START (msg_count_start)
#define MSG_COUNT_END (msg_count_end)

int steps;
#define STEPS (steps)

#define MSG_SIZE_STEP ( (((double)MSG_SIZE_END - (double)MSG_SIZE_START) / ((double)STEPS - 1.0)))
#define MSG_COUNT_STEP ( (((double)MSG_COUNT_END - (double)MSG_COUNT_START) / ((double)STEPS - 1.0)))

#define MSG_SIZE_TEST(i) (MSG_SIZE_START + (MSG_SIZE_STEP * i))
#define MSG_COUNT_TEST(i) (MSG_COUNT_START + (MSG_COUNT_STEP * i))

int warmup_count;
#define WARMUP_COUNT (warmup_count)

char *jobid;

void
read_args(int argc, char **argv)
{
    // Expect 6 args, otherwise fail
    if (argc != 7)
    {
        printf("Got %d args\n", argc);
        printf("Requires start size, end size, start count, end count, steps and warmup\n");
        exit(1);
    }

    msg_size_start = atoi(argv[1]);
    msg_size_end = atoi(argv[2]);

    msg_count_start = atoi(argv[3]);
    msg_count_end = atoi(argv[4]);

    steps = atoi(argv[5]);
    warmup_count = atoi(argv[6]);
}

int
main(int argc, char **argv)
{
    MPI_Init(&argc, &argv);

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    jobid = getenv("SLURM_JOB_ID");
    if (jobid == NULL)
    {
	jobid = getenv("PBS_JOBID");
	if (jobid == NULL) {
	    jobid = "0";
	}
    }

    read_args(argc, argv);

    if (rank == 0)
    {
        printf("About to run %d tests, each with %d steps between %d nodes\n", (((size*size) / 2 - 1)), STEPS,  size);
        printf("Starting with %d messages of %ld bytes per message\n", MSG_COUNT_START, (long) MSG_SIZE_START * 8);
        printf("Each step message count changes by %f\n", MSG_COUNT_STEP);
        printf("And message size by %f\n", MSG_SIZE_STEP);
    }

    for(int i = 0; i < size; i++)
    {
        for (int j = i + 1; j < size; j++)
        {
            MPI_Comm test_comm;
            if (rank == i)
            {
                // Initiator side
                MPI_Comm_split(MPI_COMM_WORLD, 1, rank, &test_comm);

                int test_size;
                MPI_Comm_size(test_comm, &test_size);

                // Open file
                char filename[100];
                sprintf(filename, "output/job-%s-test-%d-%d-result.tsv", jobid, i, j);
                FILE *output = fopen(filename, "w");

                // Do running SSE
                double meanX = 0;
                double meanY = 0;
                double varX = 0;
                double covXY = 0; 
                double n = 0;

                // For number of STEPS
                for (int k = 0; k < STEPS; k++)
                {
                    // Allocate data to be sent
                    double *data = calloc(MSG_SIZE_TEST(k), sizeof(double));

                    for (int l = 0; l < WARMUP_COUNT; l++)
                    {
                        // SEND
                        MPI_Ssend(data, MSG_SIZE_TEST(k), MPI_DOUBLE, 1, 0, test_comm);
                        // RECEIVE
                        //MPI_Recv(data, MSG_SIZE_TEST(k), MPI_DOUBLE, 1, 0, test_comm, MPI_STATUS_IGNORE);
                    }

                    // Barrier
                    MPI_Barrier(test_comm);
                        
                    // Clock
                    double t1 = MPI_Wtime();
                    //timespec_get(&start, TIME_UTC);
                    //clock_gettime(CLOCK_MONOTONIC, &start);
                    //gettimeofday(&start_clock, NULL);
                    // For each test
                    for (int l = 0; l < MSG_COUNT_TEST(k); l++)
                    {
                        // SEND
                        MPI_Ssend(data, MSG_SIZE_TEST(k), MPI_DOUBLE, 1, 0, test_comm);
                        // RECEIVE
                        //MPI_Recv(data, MSG_SIZE_TEST(k), MPI_DOUBLE, 1, 0, test_comm, MPI_STATUS_IGNORE);
                    }
                    // CLOCK
                    double t2 = MPI_Wtime();
                    //timespec_get(&end, TIME_UTC);
                    //clock_gettime(CLOCK_MONOTONIC, &end);

                    //double one_way_time = (((double)((end.tv_sec - start.tv_sec) * 1e9) + ((double)end.tv_nsec - (double)start.tv_nsec)) / 2.0);
                    double one_way_time = (t2 - t1) / (MSG_COUNT_TEST(k));
                    long actual_size = ((long) MSG_SIZE_TEST(k)) * 8;
                    // Write diff / 2 and size * 8 to file
                    fprintf(output, "%ld\t%.12f\n", actual_size, one_way_time);

                    // Update the SSE
                    n += 1;
                    double dX = actual_size - meanX;
                    double dY = one_way_time - meanY;
                    varX += (((n-1)/n)*dX*dX - varX)/n;
                    covXY += (((n-1)/n)*dX*dY - covXY)/n;
                    meanX += dX/n;
                    meanY += dY/n;

                    // Free memory
                    free(data);
                }

                double a = covXY / varX;
                double b = meanY - a*meanX;

                printf("SSE on %d-%d gave y= %.12fx + %.12f\n", i, j, a, b);

                fclose(output);

                MPI_Comm_free(&test_comm);
            }
            else if (rank == j)
            {
                // Participant side
                MPI_Comm_split(MPI_COMM_WORLD, 1, rank, &test_comm);

                // For number of STEPS
                for (int k = 0; k < STEPS; k++)
                {
                    // Allocate data to be sent
                    double *data = calloc(MSG_SIZE_TEST(k), sizeof(double));

                    // Prime for test
                    for (int l = 0; l < WARMUP_COUNT; l++)
                    {
                        // RECV
                        MPI_Recv(data, MSG_SIZE_TEST(k), MPI_DOUBLE, 0, 0, test_comm, MPI_STATUS_IGNORE);
                        // SEND
                        //MPI_Ssend(data, MSG_SIZE_TEST(k), MPI_DOUBLE, 0, 0, test_comm);
                    }

                    // Barrier
                    MPI_Barrier(test_comm);

                    // For each test
                    for (int l = 0; l < MSG_COUNT_TEST(k); l++)
                    {
                        // RECV
                        MPI_Recv(data, MSG_SIZE_TEST(k), MPI_DOUBLE, 0, 0, test_comm, MPI_STATUS_IGNORE);
                        // SEND
                        //MPI_Ssend(data, MSG_SIZE_TEST(k), MPI_DOUBLE, 0, 0, test_comm);
                    }

                    // Free memory
                    free(data);
                } 

                MPI_Comm_free(&test_comm);
            }
            else {
                // Not participating
                MPI_Comm_split(MPI_COMM_WORLD, MPI_UNDEFINED, rank, &test_comm);
            }
        }

        MPI_Barrier(MPI_COMM_WORLD);
    }


    MPI_Finalize();
}
