/**
 * 
 * This file contains helpers to aid the indexing from a lattice to a regular grid, ie D2Q6 to D2Q9 lattice mapping.
 * 
 * 
 * */
#pragma once
constexpr int NOW = 0;
constexpr int NEXT = 1;

static const int offsets[2][6][2] =          /* Even/odd neighbor coordinate shifts */
{
    { {0,1}, {1,0}, {1,-1}, {0,-1}, {-1,-1}, {-1,0} },
    { {0,1}, {1,1}, { 1,0}, {0,-1}, {-1, 0}, {-1,1} }
};

/**
 * @brief Return the offset of the local neightbor.
 * 
 * @param row The global row of the point
 * @param col The global column of the point
 * @param dir The direction 
 * @return int 
 */
inline int local_neighbor_x_offset(int row, int dir)
{
    if ((row-1) % 2)    return ((offsets[1][dir][1]));
    else                return ((offsets[0][dir][1]));
}

/**
 * @brief Return the offset of the local neightbor row.
 * 
 * @param row The global row of the point
 * @param dir The direction 
 * @return int 
 */
inline int local_neighbor_y_offset (int row, int dir) {
    if ((row-1) % 2)    return ((offsets[1][dir][0]));
    else                return ((offsets[0][dir][0]));
}
