#include "lattice.hpp"
#include <iomanip>

std::ostream& operator<<(std::ostream& os, const Lpoint& lp) {
    // Format is horiz vert, then density
    os << std::scientific <<std::setprecision(8) << (int) lp.type << lp.velocity[1] << " " << lp.velocity[0] 
        << " " << lp.density[0][0] << " " << lp.density[0][1] << " " << lp.density[1][0] << " " << lp.density[1][1]
        << " " << lp.density[2][0] << " " << lp.density[2][1] << " " << lp.density[3][0] << " " << lp.density[3][1] 
        << " " << lp.density[4][0] << " " << lp.density[4][1] << " " << lp.density[5][0] << " " << lp.density[5][1];
    return os;
}

std::ostream& write_lattice_to_stream(std::ostream& os, const Lattice& lat) {
    // SLightly modified from orginal lbm
    for (int row = 0; row < lat.height; row++) {
        for (int col = 0; col < lat.width; col++) {
            float norm = sqrt(lat.at(row, col).velocity[0] * lat.at(row, col).velocity[0] + lat.at(row, col).velocity[1] * lat.at(row, col).velocity[1]);
            if (norm < 1e-8) {
                norm = 1e-8;
            }
            os << std::fixed << std::setprecision(1) << std::setw(5) << (float)(col + (row&1) * 0.5) << " " << std::setw(5) << (float) row << " " 
                << std::scientific << std::setprecision(4) << " " << lat.at(row, col).velocity[1] / norm << " " << lat.at(row, col).velocity[0] / norm << '\n';
        }
    }
    return os;
}
/* From old:
void
write_velocities ( const char *filename )
{
    FILE *out = fopen ( filename, "w" );
    for ( int y=0; y<L_HEIGHT; y++ )
        for ( int x=0; x<L_WIDTH; x++ )
        {
            float norm = sqrt( global_lattice[GLI(x, y)].velocity[1] * global_lattice[GLI(x, y)].velocity[1] + global_lattice[GLI(x, y)].velocity[0] * global_lattice[GLI(x, y)].velocity[0]);
            if ( norm < 1e-8 )
                norm = 1e-8;
            // Coordinates: horiz vert, Vectors: horiz vert
            fprintf ( out, "%f %f %.4e %.4e\n",
                (float)(x + (y&1) * 0.5),
                (float)y,
                global_lattice[GLI(x, y)].velocity[1]/norm,
                global_lattice[GLI(x, y)].velocity[0]/norm
            );
        }
    fclose ( out );
}

*/

// This assumes the lattice is not a sublattice of another. 
// Of needed, this must be augmented with an input section
// which describes this lattice location in another.
void initialize_even_distribution(Lattice &l) {
    for (int row = 0; row < l.height; row++) {
        for (int col = 0; col < l.width; col++) {
            for (int i = 0; i < 6; i++) {   
                if (l.at(row, col).type == Point_type::Fluid) {
                    l.at(row,col).density[i][NOW] = l.at(row, col).density[i][NEXT] = 1.0/6.0;
                }
                else {
                    int n_x = col + local_neighbor_x_offset(row, i);
                    int n_y = row + local_neighbor_y_offset(row ,i);

                    // Is neighbor solid
                    if (l.at(n_y, n_x).type == Point_type::Fluid) {
                        l.at(row, col).density[(i+3)%6][NOW] = l.at(row, col).density[i][NEXT] = 1.0/6.0; // No
                    } else {
                        l.at(row, col).density[(i+3)%6][NOW] = l.at(row, col).density[i][NEXT] = 0; // No
                    }
                }
            }
        }
    }
}