#include <cstdlib>

#include "lbm_kernel.hpp"

void propagate(Raster task, Section global_section) {

#ifdef DATA_P_SEC
    #pragma omp parallel for
#endif // DATA_P_SEC

#ifdef TASK_P
    #pragma omp taskloop
#endif // TASK_P
    for(int y = -1; y < task.lattice->height + 1; y++) {
        for(int x = -1; x < task.lattice->width + 1; x++) {
            for(int i = 0; i < 6; i++) {
                int n_x = x + local_neighbor_x_offset(task.global_view->to_global_row(y), i);
                int n_y = y + local_neighbor_y_offset(task.global_view->to_global_row(y), i);

                // Skip if row or col is out of the global lattice, ie invalid border 
                // -> Everything is now valid border
                // if (task.global_view->to_global_row(y) < 0 
                //     || task.global_view->to_global_col(x) < 0
                //     || task.global_view->to_global_row(y) >= global_section.height 
                //     || task.global_view->to_global_col(x) >= global_section.width) 
                // {
                //     continue;
                // }

                // Skip if n_y or n_x is out of bounds
                if  (n_y < -1 || n_x < -1 || n_y >= task.lattice->height + 1 || n_x >= task.lattice->width + 1){
                    continue;
                }

                task.lattice->at(n_y, n_x).density[i][NOW] = task.lattice->at(y, x).density[i][NEXT];
            }
        }
    }
}

void collide2(Raster task, std::function<void(Section&, int, int, int, float&)> apply_force) 
{

#ifdef DATA_P_SEC
    #pragma omp parallel for
#endif // DATA_P_SEC

#ifdef TASK_P
    #pragma omp taskloop
#endif // TASK_P
    for (int row = 0; row < task.global_view->height; row++) {
        for (int col = 0; col < task.global_view->width; col++) {
            float rho = 0.0;
            float uc = 0.0;

            task.lattice->at(row, col).velocity[NOW] = task.lattice->at(row, col).velocity[NEXT] = 0.0;

            if ( task.lattice->at(row, col).type == Point_type::Fluid ) {
                for ( int i=0; i<6; i++ )
                {
                    rho += task.lattice->at(row, col).density[i][NOW];
                    task.lattice->at(row, col).velocity[NOW] += c[i][0] * task.lattice->at(row, col).density[i][NOW];
                    task.lattice->at(row, col).velocity[NEXT] += c[i][1] * task.lattice->at(row, col).density[i][NOW];
                }
                /* rho*u = sum_i( Ni*ci ), so divide by rho to find u: */
                task.lattice->at(row, col).velocity[NOW] /= rho;
                task.lattice->at(row, col).velocity[NEXT] /= rho;
            }

            for ( int i=0; i<6; i++ )
            {
                float
                    qi_uaub,
                    N_eq,
                    delta_N;
                qi_uaub =
                    ( c[i][1] * c[i][1] - 0.5 ) * task.lattice->at(row, col).velocity[NEXT] * task.lattice->at(row, col).velocity[NEXT] +
                    ( c[i][1] * c[i][0]       ) * task.lattice->at(row, col).velocity[NEXT] * task.lattice->at(row, col).velocity[NOW] +
                    ( c[i][0] * c[i][1]       ) * task.lattice->at(row, col).velocity[NOW] * task.lattice->at(row, col).velocity[NEXT] +
                    ( c[i][0] * c[i][0] - 0.5 ) * task.lattice->at(row, col).velocity[NOW] * task.lattice->at(row, col).velocity[NOW];
                uc = task.lattice->at(row, col).velocity[NOW] * c[i][0] + task.lattice->at(row, col).velocity[NEXT] * c[i][1];

                // Equilibrium, difference
                N_eq = ( rho / 6.0 ) * ( 1.0 + 2.0 * uc + 4.0 * qi_uaub );
                delta_N = LAMBDA * ( task.lattice->at(row, col).density[i][NOW] - N_eq );

                // Apply external force in the channel
                apply_force((*task.global_view), row, col, i, delta_N);

                // Reflections at ghosts
                if( task.lattice->at(row, col).type == Point_type::Fluid ){
                    task.lattice->at(row, col).density[i][NEXT] = task.lattice->at(row, col).density[i][NOW] + delta_N;
                }
                else {
                    task.lattice->at(row, col).density[(i+3)%6][NEXT] = task.lattice->at(row, col).density[i][NOW];
                }
            }
        }
    }
}


void channel_force(Section& section, int row, int col, int i, float& delta_N) {
    if ( section.to_global_row(row) >=1 && section.to_global_row(row) < CHANNEL ) {
        delta_N += (1.0/3.0) * (c[i][0]*force[0] + c[i][1]*force[1]);
    }
}