#pragma once
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <sstream>

#include <mpi.h>

#include "raster.hpp"
#include "lbm_kernel.hpp"
#include "part_grid.hpp"
#include "part_rec.hpp"
#include "exch_plan.hpp"
#include "scheduling.hpp"
#include "pbm.h"
#include "mpi_types.hpp"
#include "debug.hpp"


extern MPI_Datatype section_type;
extern MPI_Datatype exchange_type;
extern MPI_Datatype lpoint_type;

int init_mpi_types();
int free_mpi_types();
void send_exchange_struct(const Section_exchange& exch, int rank, int tag);
Section_exchange recv_exchange_struct(int rank, int tag);

void border_exchange(   std::vector<Section>& sections, 
                        std::vector<Lattice> &lattices, 
                        std::vector<Section_exchange> &exchanges,
                        std::vector<int> &section_assignment,
                        Lattice_type_factory &lat_mpi_type_fac,
                        int rank);

void gather_sections(   std::vector<std::vector<int>>& rank_assignment,
                        std::vector<Section>& all_sections,
                        Lattice& global_lattice,
                        std::vector<int>& section_ids,
                        std::vector<Section>& sections, 
                        std::vector<Lattice> &lattices,
                        std::vector<int> &section_assignment,
                        Lattice_type_factory &lat_mpi_type_fac,
                        int rank, int size);

void receive_initial_scatter(
                        int &section_count, 
                        std::vector<int> &section_ids, 
                        std::vector<Section> &sections, 
                        std::vector<Section_exchange> &exchanges, 
                        std::vector<Lattice> &lattices, 
                        Lattice_type_factory &lat_mpi_type_fac,
                        int rank, int size);

void send_initial_scatter(
                        int i, 
                        std::vector<std::vector<int>>& rank_assignment, 
                        Lattice& global_lattice, 
                        std::vector<Section_exchange>& all_exchanges, 
                        std::vector<Section>& all_sections, 
                        Lattice_type_factory &lat_mpi_type_fac);