#include "mpi_types.hpp"

Lattice_type_factory::~Lattice_type_factory() {
    for (auto pair: subsec_types) {
        MPI_Type_free(&pair.second);
    }
    for (auto pair: local_types) {
        MPI_Type_free(&pair.second);
    }
}

MPI_Datatype Lattice_type_factory::subsec_type(const Lattice& lat, const Section& sec, int border) {
    int data_width = lat.data_width;
    int act_height = sec.height + 2*border;
    int act_width = sec.width + 2*border;

    auto existing = subsec_types.find(map_key(data_width, act_height, act_width));
    if (existing != subsec_types.end()) {
        // Found, return it
        return existing->second;
    }

    MPI_Datatype type;
    MPI_Type_vector(act_height, act_width, data_width, lpoint_type, &type);
    MPI_Type_commit(&type);

    subsec_types[map_key(data_width, act_height, act_width)] = type;

    return type;
}

MPI_Datatype init_section_type (){
    MPI_Datatype temp;
    MPI_Datatype section_type;
    // Section type
    const int block_lengths[] = {1, 1, 1, 1, 1, 1}; 
    const MPI_Aint displacements[] = {
        offsetof(Section, tl_row),
        offsetof(Section, tl_col),
        offsetof(Section, br_row),
        offsetof(Section, br_col),
        offsetof(Section, height),
        offsetof(Section, width)
    };
    const MPI_Datatype datatypes[] = {MPI_INT, MPI_INT, MPI_INT, MPI_INT, MPI_INT,MPI_INT};
    MPI_Type_create_struct(6, block_lengths, displacements, datatypes, &temp);
    MPI_Type_create_resized(temp, 0, sizeof(Section), &section_type);

    return section_type;
}

MPI_Datatype init_exchange_type() {
    MPI_Datatype section_type = init_section_type();
    MPI_Type_commit(&section_type);

    MPI_Datatype temp;
    MPI_Datatype exchange_type; 
    const int exch_blk_len[] = {1, 1, 1};
    const MPI_Aint exch_displ[] = {
        offsetof(Exchange, section),
        offsetof(Exchange, where),
        offsetof(Exchange, tag)
    };
    const MPI_Datatype exch_datatypes[] = {
        section_type,
        MPI_INT,
        MPI_INT
    };

    MPI_Type_create_struct(3, exch_blk_len, exch_displ, exch_datatypes, &temp);
    MPI_Type_create_resized(temp, 0, sizeof(Exchange), &exchange_type);

    return exchange_type;
}

MPI_Datatype init_point_type() {
    MPI_Datatype temp;
    MPI_Datatype lpoint_type;
    const int pt_blk_len[] = {1, 2, 12};
    const MPI_Aint pt_displ[] = {
        offsetof(Lpoint, type),
        offsetof(Lpoint, velocity),
        offsetof(Lpoint, density)
    };
    const MPI_Datatype pt_datatypes[] = {
        MPI_INT,
        MPI_FLOAT,
        MPI_FLOAT
    };
    MPI_Type_create_struct(3, pt_blk_len, pt_displ, pt_datatypes, &temp);
    MPI_Type_create_resized(temp, 0, sizeof(Lpoint), &lpoint_type);
    
    return lpoint_type;
}
