#include <vector>
#include "raster.hpp"
#include <algorithm>
#include "part_grid.hpp"

std::vector<Section> generate_grid(std::pair<int, int> matrix_size, std::pair<int, int> cell_size) {
    // First create all partitions
    std::vector<Section> grids;

    int current_row = 0;
    while (current_row < matrix_size.first) {
        int current_column = 0;
        while(current_column < matrix_size.second) {
            int height = std::min(matrix_size.first - current_row, cell_size.first);
            int width = std::min(matrix_size.second - current_column, cell_size.second);

            grids.push_back(Section(current_row, current_column, height, width));

            current_column += cell_size.second;
        }
        current_row += cell_size.first;
    }
    return grids;
}

struct section_is_empty
{
    Lattice* lattice;
    section_is_empty(Lattice* lattice): lattice(lattice) {}

    bool operator()(Section& sec) {
        for (int row = 0; row < sec.height; row++) {
            for (int col = 0; col < sec.width; col++) {
                if (lattice->at(sec.to_global_row(row), sec.to_global_col(col)).type == Point_type::Fluid) {
                    return false;
                }
            }
        }
        return true;
    }  
};

std::vector<Section> partition_grid(Lattice& lattice, std::pair<int, int> dims, bool filterEmpty) {    
    // Partitin into fixed sized cells/subgrids
    auto cells = generate_grid({lattice.height, lattice.width}, dims);

    // Filter out empty cells
    if (filterEmpty) {
        cells.erase(std::remove_if(cells.begin(), cells.end(), section_is_empty(&lattice)), cells.end());
    }

    return cells;
}