#include "part_rec.hpp"


struct Recursive_splitter 
{
    Lattice* lat;
    std::array<int,2> stop_size;
    std::vector<int> sums; // Used in balance cal, preallocated once, makes this not threadsafe

    Recursive_splitter(Lattice* lat, std::array<int, 2> stop_size): lat(lat), stop_size(stop_size), sums(std::max(lat->height, lat->width), 0) {}

    int determine_hori_equil(Section& sec);
    int determine_vert_equil(Section& sec);

    std::vector<Section> split_rec(Section& sec, bool horizontal = true);
};

int Recursive_splitter::determine_hori_equil(Section& sec) {
    // Brute force, scan the entire thing, then back up half way
    
    // This case is top to bottom scan
    int sum = 0;

    for(int row = 0; row < sec.height; row++) {
        for (int col = 0; col < sec.width; col++) {
            sum += 1 - static_cast<int>(lat->at(sec.to_global_row(row), sec.to_global_col(col)).type);
            sums[row] += 1 - static_cast<int>(lat->at(sec.to_global_row(row), sec.to_global_col(col)).type);
        }
    }

    if (sum == 0) {
        return -1;
    }

    // Now scan again until we have half
    float half = static_cast<float>(sum) / 2;
    int running = 0;
    
    int row;
    for(row = 0; row < sec.height; row++)
    {
        running += sums[row];

        if (running > half) {
            break;
        }
    }

    return row;
}

int Recursive_splitter::determine_vert_equil(Section& sec) {
    // Brute force, scan the entire thing, then back up half way
    
    // This case is left to rigt scan
    int sum = 0;

    for (int col = 0; col < sec.width; col++) {
        for(int row = 0; row < sec.height; row++) {
            // Each point is 0 - fluid og 1 - solid. Here we interpret that as bool ,and flip it to get a useful sum to divide.
            sum += 1 - static_cast<int>(lat->at(sec.to_global_row(row), sec.to_global_col(col)).type);
            sums[col] += 1 - static_cast<int>(lat->at(sec.to_global_row(row), sec.to_global_col(col)).type);
        }
    }

    if (sum == 0) {
        return -1;
    }

    // Now scan again until we have half
    float half = static_cast<float>(sum) / 2;
    int running = 0;
    
    int col;
    for(col = 0; col < sec.width; col++)
    {
        running += sums[col];

        if (running > half) {
            break;
        }
    }

    return col;
}

std::array<Section,2> split_hori(Section& sec, int equil) {
    int middle = sec.height / 2;
    return {Section(sec.tl_row, sec.tl_col, middle, sec.width), Section(sec.tl_row + middle, sec.tl_col, sec.height - middle, sec.width)};
}

std::array<Section,2> split_vert(Section& sec, int equil) {
    int middle = sec.width / 2;
    return {Section(sec.tl_row, sec.tl_col, sec.height, middle), Section(sec.tl_row, sec.tl_col + middle, sec.height, sec.width - middle)};
}

std::vector<Section> Recursive_splitter::split_rec(Section& sec, bool horizontal) 
{
    // Check if it is already small enough
    if (sec.height <= stop_size[0] && sec.width <= stop_size[1]) { // 
        return {sec};
    } else if (horizontal ? sec.height <= stop_size[0] : sec.width <= stop_size[1]) {
        return split_rec(sec, !horizontal);
    }

    // Find middle in alternating directions, 
    // horizontal means move from top down
    // splitting along the horizontal axis
    int equil = horizontal ? determine_hori_equil(sec) : determine_vert_equil(sec);

    // If split point is -1, it's empty and we return empty
    if (equil == -1) {
        return std::vector<Section>();
    }

    // Split into two using the middle point
    std::array<Section, 2> newSections = horizontal ? split_hori(sec, equil) : split_vert(sec, equil);

    // Recurse into both, returning the union of the two
    auto newSectionsOne = split_rec(newSections[0], !horizontal);
    auto newSectionTwo = split_rec(newSections[1], !horizontal);
    newSectionsOne.insert(newSectionsOne.end(), std::make_move_iterator(newSectionTwo.begin()), std::make_move_iterator(newSectionTwo.end()));
    return newSectionsOne;
}

std::vector<Section> partition_rec(Lattice& lattice, std::pair<int, int> stop_dims) 
{
    // Make a recursive splitter
    Recursive_splitter splitter(&lattice, {stop_dims.first, stop_dims.second});

    // Make an initial section that cover the entire lattice
    Section global(0, 0, lattice.height, lattice.width);
    
    // Split
    return splitter.split_rec(global);
}