/**
 * @file pbm.h
 * @author your name (you@domain.com)
 * @brief Implementation of reading and writing lattice to PBM image format https://en.wikipedia.org/wiki/Netpbm
 * @version 0.1
 * @date 2020-02-03
 * 
 * @copyright Copyright (c) 2020
 * 
 */
#pragma once
#include "raster.hpp"
#include <string>
#include <iostream>
#include <fstream>

static inline Lattice load_from_stream(std::istream& stream) {
    using namespace std;
    string magic_number, width, height;
    stream >> magic_number >> width >> height;

    Lattice lattice(atoi(height.c_str()), atoi(width.c_str()));
    
    char c;
    int i = 0;
    while (stream >> c)
    {
        if (c == '0') {
            lattice.at(i / lattice.width, i % lattice.width).type = Point_type::Fluid;
            i++;
        }
        else if (c == '1') {
            lattice.at(i / lattice.width, i % lattice.width).type = Point_type::Solid;
            i++;
        } 
    }
    
    return lattice;
}

static inline Lattice load_from_image(std::string filename) {
    using namespace std;
    ifstream image_file(filename.c_str());

    if(!image_file) {
        cerr << "Error opening image file";
        exit(EXIT_FAILURE);
    }

    return load_from_stream(image_file);
}