#include "scheduling.hpp"

// We need to sort based on load -> Assume load is only dependent on size
struct Section_load {
    int section;
    double load;

    Section_load(int section, double load): section(section), load(load) {}
    bool operator<(const Section_load& other) {return load > other.load;} // We flip the operator so that the sorting will be decending
};

struct Work_assignment {
    std::vector<int> sections;
    double total_work = 0;

    void add_section(const Section_load& sec) {
        sections.push_back(sec.section);
        total_work += sec.load;
    }

    bool operator<(const Work_assignment& other) {return total_work <= other.total_work;}
};

std::vector<std::vector<int>> schedule_greedy(const std::vector<Section>& sections, int ranks, std::function<double(const Section&)> cost_function){
    std::vector<Section_load> load;
    
    // Determine load
    for (int i = 0; i < sections.size(); i++) {
        load.push_back(Section_load(i, cost_function(sections[i])));
    }

    // Sort by load
    std::sort(load.begin(), load.end());

    // Initialize the work plan with empty assignments per rank
    std::vector<Work_assignment> work_plan(ranks);

    // Now we make the plan
    for (int i = 0; i < load.size(); i++) {
        // We take the next section and assign it to the most idle rank
        work_plan[0].add_section(load[i]);

        // Sort the workplan to ensure the most idle is always first
        std::sort(work_plan.begin(), work_plan.end()); 
    }

    // Move it into return format
    std::vector<std::vector<int>> return_value;
    return_value.reserve(ranks);
    for (int i = 0; i < ranks; i++){
        return_value.push_back(std::move(work_plan[i].sections)); // Use move to avoid copy
    }

    return return_value;
}

std::vector<std::vector<int>> schedule_round_robin(const std::vector<Section>& sections, int ranks) {
    std::vector<std::vector<int>> work_plan(ranks);
    for (int i = 0; i < sections.size(); i++) {
        work_plan[i % ranks].push_back(i);
    }
    return work_plan;
}
