#pragma once

#include <map>
#include <memory>
#include "lattice.hpp"
#include <iostream>

#include "doctest.hpp"

struct Section {
    int tl_row,
        tl_col,
        br_row,
        br_col,
        height,
        width;
    
    Section(){}
    Section(int tl_row, int tl_col, int height, int width): tl_row(tl_row), tl_col(tl_col), br_row(tl_row + height - 1), 
                                                            br_col(tl_col + width - 1), height(height), width(width) {
                                                                CHECK_GE(height, 0);
                                                                CHECK_GE(width, 0);
                                                            }

    Section(std::pair<int,int> tl, int height, int width): Section(tl.first, tl.second, height, width) {}

    Section(std::pair<int,int> tl, std::pair<int,int> br):  tl_row{tl.first}, tl_col{tl.second}, 
                                                            br_row{br.first}, br_col{br.second}, 
                                                            height{br_row - tl_row + 1}, 
                                                            width{br_col - tl_col + 1} {
                                                                CHECK_GE(height, 0);
                                                                CHECK_GE(width, 0);
                                                            }

    inline int to_global_row(int row) const { return row + tl_row; }
    inline int to_global_col(int col) const { return col + tl_col; }
    inline int tgr(int row) const { return row + tl_row; }
    inline int tgc(int col) const { return col + tl_col; }

    inline int from_global_row(int row) const { return row - tl_row; }
    inline int from_global_col(int col) const { return col - tl_col; }
    inline int fgr(int row) const { return row - tl_row; }
    inline int fgc(int col) const { return col - tl_col; }
};

std::ostream& operator<<(std::ostream& os, const Section& sec);

bool operator==(const Section& lhs, const Section& rhs);